function projectedImage = projectImage(frames, sourceFrameIndex, referenceFrameIndex, pastHomographies, xrange, yrange, overlapThreshold, errorThreshold, numKeyframes, checkAllKeyframes)
    
    % Input: 
    % - frames: 4D array of frames
    % - sourceFrameIndex: index of the frame to be projected
    % - referenceFrameIndex: index of the frame to be projected to
    % - pastHomographies: 2D cell array caching previously computed
    % homographies from every frame to every other frame
    % - xrange, yrange: dimensions of the output image
    % - overlapThreshold: sufficient number of pixels overlapping between 
    % projected source and reference frames to ensure good homography
    % - errorThreshold: acceptable error for good homography
    % - numKeyframes: number of equidistant keyframes between source and
    % reference frame to be visited in search of better homography
    % - checkAllKeyframes: 0 if algorithms breaks after first better 
    % homography is found, 1 if all keyframes are to be visited
    
    % Output: 
    % - bestProjectedImage: source frame optimally projected onto reference
    % frame using reestimation of homography based on closest-frame search
    % and using closest-frame homography as  
    
    SetDefaultValue(5, 'xrange', [-651 980]);
    SetDefaultValue(6, 'yrange', [-51 460]);
    SetDefaultValue(7, 'overlapThreshold', 40000);
    SetDefaultValue(8, 'errorThreshold', 4e-4);
    SetDefaultValue(9, 'numKeyframes', 3);
    SetDefaultValue(10, 'checkAllKeyframes', 0);
    
    numFrames = size(frames, 4);
    
    [~, referenceTransform] = transformImage(frames, referenceFrameIndex, referenceFrameIndex, pastHomographies, xrange, yrange);
    [~, sourceTransform] = transformImage(frames, sourceFrameIndex, referenceFrameIndex, pastHomographies, xrange, yrange);
    [~, err] = computeOverlap(sourceTransform, referenceTransform, overlapThreshold);
    projectedImage = sourceTransform;
        
    if err > errorThreshold
        disp('Finding better homography...');
        increment = floor(((referenceFrameIndex - sourceFrameIndex) - 1) / (numKeyframes + 1));
        keyframeIndex = sourceFrameIndex + increment; % frame being used to find better homography from source to reference
        found = 0;
        counter = 0;
        bestHomography = eye(3); % initialize H as identity
        
        while counter < numKeyframes && keyframeIndex < numFrames && keyframeIndex > 0
            
            % compute homography and projected image from keyframe to
            % reference frame
            [H2, keyframeTransform] = transformImage(frames, keyframeIndex, referenceFrameIndex, pastHomographies, xrange, yrange);
            [a, error1] = computeOverlap(keyframeTransform, referenceTransform, overlapThreshold);
            
            % compute homography and projected image from source frame to
            % keyframe (new reference = keyframe)
            [~, keyframeToKeyframeTransform] = transformImage(frames, keyframeIndex, keyframeIndex, pastHomographies, xrange, yrange);
            [H1, sourceToKeyframeTransform] = transformImage(frames, sourceFrameIndex, keyframeIndex, pastHomographies, xrange, yrange);
            [b, error2] = computeOverlap(sourceToKeyframeTransform, keyframeToKeyframeTransform, overlapThreshold);
            
            sufficientOverlap = (a && b);
            
            if (sufficientOverlap && max(error1, error2) < err)
                found = 1;
                bestHomography = H1 * H2;
                if (~checkAllKeyframes)
                    break
                end
            end
            
            keyframeIndex = keyframeIndex + increment;
            counter = counter + 1;
        end
        
        if (found)
            disp({'Found better homography'});
            pastHomographies{sourceFrameIndex, referenceFrameIndex} = bestHomography;
            T = maketform('projective', bestHomography');
            projectedImage = imtransform(frames(:,:,:,sourceFrameIndex), T, 'XData', xrange, 'YData', yrange);
        end
    end
    
end